unit LiveIni;

interface

uses
  System.Generics.Collections;

type
  TLiveIni = class(TObject)
    private
      type
        TSection = TDictionary<string, string>;
        TIni = TDictionary<string, TSection>;
      var
        FIni: TIni;
        FChanged: Boolean;
      function GetValue(const Section, Key: string): string;
      procedure SetValue(const Section, Key, Value: string);
      function ValueToStr(const Section, Line: string): string;
    public
      constructor Create();
      procedure Load(const InFile: string);
      procedure Save(const OutFile: string);
      destructor Destroy(); override;
      property Changed: Boolean read FChanged;
      property Values[const Section, Key: string]: string read GetValue write SetValue; default;
  end;

implementation

uses
  System.Classes, System.RegularExpressionsCore, System.StrUtils, System.SysUtils;

var
  CommentRegEx: TPerlRegEx;
  SectionRegEx: TPerlRegEx;
  KeyValueRegEx: TPerlRegEx;
  TokenRegEx: TPerlRegEx;

constructor TLiveIni.Create();
begin
  inherited Create;
  FIni := TIni.Create; // Create FIni
  FChanged := False;   // Set unchanged
end;

procedure TLiveIni.Load(const InFile: string);
var
  t: TStringList;
  n, k, v: string;
  i: Integer;
  s: TSection;
begin
  for n in FIni.Keys do begin                                     // For each Section
    FIni.TryGetValue(n, s);                                       // Get Section
    s.Free;                                                       // Free Section
  end;
  FIni.Free;                                                      // Free FIni
  FIni := TIni.Create;                                            // Create FIni
  t := TStringList.Create(True);                                  // Create string list for file Text
  try
    t.LoadFromFile(InFile);                                       // Load Text
    for i := 0 to (t.Count - 1) do begin                          // For each Text line
      CommentRegEx.Subject := t[i];                               // Check for Comment
      if CommentRegEx.Match then                                  // Comment
        t[i] := AnsiReplaceStr(t[i], CommentRegEx.Groups[1], ''); // Remove Comment
      SectionRegEx.Subject := t[i];                               // Check for Section name
      KeyValueRegEx.Subject := t[i];                              // Check for Key/Value pair
      if SectionRegEx.Match then                                  // Section name
        begin
          n := AnsiUpperCase(SectionRegEx.Groups[1]);             // Set current Section
          if not (FIni.ContainsKey(n)) then                       // Check for Section
            FIni.Add(n, TSection.Create);                         // Add new Section to Ini
        end
      else if KeyValueRegEx.Match then                            // Key/Value pair
        begin
          k := AnsiUpperCase(KeyValueRegEx.Groups[1]);            // Key
          v := KeyValueRegEx.Groups[2];                           // Value
          FIni[n].AddOrSetValue(k, v);                            // Add or change Key/Value pair
        end;
    end;
  finally
    t.Free;                                                       // Delete Text
  end;
  FChanged := False;                                              // Set unchanged
end;

procedure TLiveIni.Save(const OutFile: string);
var
  t: TStringList;
  SecList: TStringList;
  KeyList: TStringList;
  s, k, v: string;
  i, j: Integer;
begin
  t := TStringList.Create();                    // Create string list for file Text
  SecList := TStringList.Create();              // Create string list for Sections
  SecList.Sorted := True;                       // Section list is sorted
  for s in FIni.Keys do SecList.Append(s);      // Append Section names to list
  for i := 0 to (SecList.Count - 1) do begin    // For each Section in list (sorted)
    s := SecList[i];                            // Get Section name
    t.Append('[' + s + ']');                    // Append Section to Text list
    KeyList := TStringList.Create();            // Create Key list for Section
    KeyList.Sorted := True;                     // Key list is sorted
    for k in FIni[s].Keys do KeyList.Append(k); // Append Key names to list
    for j := 0 to (KeyList.Count - 1) do begin  // For each Key in list (sorted)
      k := KeyList[j];                          // Get Key name
      FIni[s].TryGetValue(k, v);                // Get Vlaue
      t.Append(k + '=' + v);                    // Append Key/Value to Text list
    end;
    KeyList.Free;                               // Free Key list
    t.Append('');                               // Append blank line to Text list
  end;
  SecList.Free;                                 // Free Section list
  try
    t.SaveToFile(OutFile);                      // Save Text to file
  finally
    t.Free;                                     // Free Text list
  end;
  FChanged := False;                            // Set unchanged
end;

destructor TLiveIni.Destroy();
var
  n: string;
  s: TSection;
begin
  for n in FIni.Keys do begin // For each Section
    FIni.TryGetValue(n, s);   // Get Section
    s.Free;                   // Free Section
  end;
  FIni.Free;                  // Free FIni
  inherited Destroy;
end;

function TLiveIni.GetValue(const Section, Key: string): string;
var
  s: TSection;
  v: string;
begin
  if FIni.TryGetValue(AnsiUpperCase(Section), s) then           // Get Section
    if s.TryGetValue(AnsiUpperCase(Key), v) then                // Get Value
      Result := ValueToStr(Section, v)                          // Return normalized Value
    else                                                        // No Key
      Result := ('[BAD KEY "' +  Section + ':' + Key + '"]')    // Return Error
  else                                                          // No Section
    Result := ('[BAD SECTION "' +  Section + ':' + Key + '"]'); // Return Error
end;

procedure TLiveIni.SetValue(const Section, Key, Value: string);
begin
  if not (FIni.ContainsKey(AnsiUpperCase(Section))) then                     // New Section?
    begin
      FIni.Add(AnsiUpperCase(Section), TSection.Create);                     // Add new Section to Ini
      FChanged := True;                                                      // Ini changed
    end;
  if not SameText(GetValue(Section, Key), Value) then                        // New Value?
    begin
      FIni[AnsiUpperCase(Section)].AddOrSetValue(AnsiUpperCase(Key), Value); // Add or change Key/Value
      FChanged := True;                                                      // Ini changed
    end;
end;

function TLiveIni.ValueToStr(const Section, Line: string): string;
var
  c, k, r, s, t, v: string;
begin
  TokenRegEx.Subject := Line;                        // Check for Token
  while TokenRegEx.Match do begin
    v := TokenRegEx.Groups[0];                       // Whole Value
    t := TokenRegEx.Groups[1];                       // Token
    c := TokenRegEx.Groups[2];                       // +/- (upper/lower case)
    s := TokenRegEx.Groups[3];                       // Section
    k := TokenRegEx.Groups[4];                       // Key
    if (s = '*') then s := Section;                  // Section = current Section
    r := Values[AnsiUpperCase(s), AnsiUpperCase(k)]; // Recursive lookup For Token Value
    if (c = '-') then r := AnsiLowerCase(r);         // - = lower case Token Value
    if (c = '+') then r := AnsiUpperCase(r);         // + = upper case Token Value
    TokenRegEx.Subject := AnsiReplaceStr(v, t, r);   // Replace Token with Token Value
  end;
  Result := TokenRegEx.Subject;                      // Return (possibly) modified Value
end;

initialization
begin
  CommentRegEx := TPerlRegEx.Create;
  CommentRegEx.RegEx := '^.*?(;.*)$';
  CommentRegEx.Study;
  SectionRegEx := TPerlRegEx.Create;
  SectionRegEx.RegEx := '^\s*\[\s*(.+?)\s*\].*$';
  SectionRegEx.Study;
  KeyValueRegEx := TPerlRegEx.Create;
  KeyValueRegEx.RegEx := '^\s*(.+?)\s*\=\s*(.*?)\s*$';
  KeyValueRegEx.Study;
  TokenRegEx := TPerlRegEx.Create;
  TokenRegEx.RegEx := '^.*(<\s*([\-\+]?)\s*(.+?)\s*\:\s*(.*?)\s*>).*$';
  TokenRegEx.Study;
end;

finalization
begin
  CommentRegEx.Free;
  SectionRegEx.Free;
  KeyValueRegEx.Free;
  TokenRegEx.Free;
end;

end.